/*
 * Copyright (c) 2011-2014, Peter Abeles. All Rights Reserved.
 *
 * This file is part of BoofCV (http://boofcv.org).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package boofcv.alg.filter.convolve.border;

import boofcv.misc.AutoTypeImage;
import boofcv.misc.CodeGeneratorUtil;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;

/**
 * Code generator for {@link ConvolveJustBorder_General}.  
 *
 * @author Peter Abeles
 */
public class GenerateConvolveJustBorder_General {

	String className = "ConvolveJustBorder_General";

	PrintStream out;

	AutoTypeImage imageOut;
	String typeKernel;
	String typeInput;
	String dataKernel;
	String sumType;

	public void createAll() throws FileNotFoundException {
		out = new PrintStream(new FileOutputStream(className + ".java"));

		printPreamble();

		createMethod(AutoTypeImage.F32);
		createMethod(AutoTypeImage.I16);
		createMethod(AutoTypeImage.S32);

		createBound(AutoTypeImage.F32);
		createBound(AutoTypeImage.I8);

		out.println("}");
	}

	public void createMethod( AutoTypeImage imageOut ) {
		this.imageOut = imageOut;

		if( imageOut.isInteger() ) {
			typeKernel = "I32";
			dataKernel = "int";
			typeInput = "ImageBorder_I32";
		} else {
			typeKernel = "F32";
			dataKernel = "float";
			typeInput = "ImageBorder_F32";
		}

		sumType = imageOut.getSumType();
		String typeCast = imageOut.getTypeCastFromSum();

		addHorizontal(typeCast);
		addVertical(typeCast);
		addConvolution(typeCast);
	}

	public void createBound( AutoTypeImage imageOut ) {
		this.imageOut = imageOut;

		if( imageOut.isInteger() ) {
			typeKernel = "I32";
			dataKernel = "int";
			typeInput = "ImageBorder_I32";
		} else {
			typeKernel = "F32";
			dataKernel = "float";
			typeInput = "ImageBorder_F32";
		}

		sumType = imageOut.getSumType();
		String typeCast = imageOut.getTypeCastFromSum();

		addConvolutionBound(typeCast);
	}

	public void printPreamble() {
		out.print(CodeGeneratorUtil.copyright);
		out.print("package boofcv.alg.filter.convolve.border;\n" +
				"\n" +
				"import boofcv.core.image.border.ImageBorder_F32;\n" +
				"import boofcv.core.image.border.ImageBorder_I32;\n" +
				"import boofcv.struct.convolve.Kernel1D_F32;\n" +
				"import boofcv.struct.convolve.Kernel1D_I32;\n" +
				"import boofcv.struct.convolve.Kernel2D_F32;\n" +
				"import boofcv.struct.convolve.Kernel2D_I32;\n" +
				"import boofcv.struct.image.*;\n" +
				"\n" +
				"/**\n" +
				" * <p>\n" +
				" * Convolves just the image's border.  How the border condition is handled is specified by the {@link boofcv.core.image.border.ImageBorder}\n" +
				" * passed in.  For 1D kernels only the horizontal or vertical borders are processed.\n" +
				" * </p>\n" +
				" * \n" +
				" * <p>\n" +
				" * WARNING: Do not modify.  Automatically generated by {@link GenerateConvolveJustBorder_General}.\n" +
				" * </p>\n" +
				" * \n" +
				" * @author Peter Abeles\n" +
				" */\n" +
				"public class ConvolveJustBorder_General {\n\n");
	}

	public void addHorizontal( String typeCast ) {
		String typeOutput = imageOut.getSingleBandName();
		String dataOutput = imageOut.getDataType();

		out.print("\tpublic static void horizontal(Kernel1D_"+typeKernel+" kernel, "+typeInput+" input, "+typeOutput+" output ) {\n" +
				"\t\tfinal "+dataOutput+"[] dataDst = output.data;\n" +
				"\t\tfinal "+dataKernel+"[] dataKer = kernel.data;\n" +
				"\n" +
				"\t\tfinal int offset = kernel.getOffset();\n" +
				"\t\tfinal int kernelWidth = kernel.getWidth();\n" +
				"\t\tfinal int width = output.getWidth();\n" +
				"\t\tfinal int height = output.getHeight();\n" +
				"\t\tfinal int borderRight = kernelWidth-offset-1;\n" +
				"\n" +
				"\t\tfor (int y = 0; y < height; y++) {\n" +
				"\t\t\tint indexDest = output.startIndex + y * output.stride;\n" +
				"\n" +
				"\t\t\tfor ( int x = 0; x < offset; x++ ) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tfor (int k = 0; k < kernelWidth; k++) {\n" +
				"\t\t\t\t\ttotal += input.get(x+k-offset,y) * dataKer[k];\n" +
				"\t\t\t\t}\n" +
				"\t\t\t\tdataDst[indexDest++] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\n" +
				"\t\t\tindexDest = output.startIndex + y * output.stride + width-borderRight;\n" +
				"\t\t\tfor ( int x = width-borderRight; x < width; x++ ) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tfor (int k = 0; k < kernelWidth; k++) {\n" +
				"\t\t\t\t\ttotal += input.get(x+k-offset,y) * dataKer[k];\n" +
				"\t\t\t\t}\n" +
				"\t\t\t\tdataDst[indexDest++] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\t\t}\n" +
				"\t}\n\n");
	}

	public void addVertical( String typeCast ) {
		String typeOutput = imageOut.getSingleBandName();
		String dataOutput = imageOut.getDataType();

		out.print("\tpublic static void vertical(Kernel1D_"+typeKernel+" kernel, "+typeInput+" input, "+typeOutput+" output ) {\n" +
				"\t\tfinal "+dataOutput+"[] dataDst = output.data;\n" +
				"\t\tfinal "+dataKernel+"[] dataKer = kernel.data;\n" +
				"\n" +
				"\t\tfinal int offset = kernel.getOffset();\n" +
				"\t\tfinal int kernelWidth = kernel.getWidth();\n" +
				"\t\tfinal int width = output.getWidth();\n" +
				"\t\tfinal int height = output.getHeight();\n" +
				"\t\tfinal int borderBottom = kernelWidth-offset-1;\n" +
				"\n" +
				"\t\tfor ( int x = 0; x < width; x++ ) {\n" +
				"\t\t\tint indexDest = output.startIndex + x;\n" +
				"\n" +
				"\t\t\tfor (int y = 0; y < offset; y++, indexDest += output.stride) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tfor (int k = 0; k < kernelWidth; k++) {\n" +
				"\t\t\t\t\ttotal += input.get(x,y+k-offset) * dataKer[k];\n" +
				"\t\t\t\t}\n" +
				"\t\t\t\tdataDst[indexDest] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\n" +
				"\t\t\tindexDest = output.startIndex + (height-borderBottom) * output.stride + x;\n" +
				"\t\t\tfor (int y = height-borderBottom; y < height; y++, indexDest += output.stride) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tfor (int k = 0; k < kernelWidth; k++ ) {\n" +
				"\t\t\t\t\ttotal += input.get(x,y+k-offset) * dataKer[k];\n" +
				"\t\t\t\t}\n" +
				"\t\t\t\tdataDst[indexDest] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\t\t}\n" +
				"\t}\n\n");
	}

	public void addConvolution( String typeCast ) {
		String typeOutput = imageOut.getSingleBandName();
		String dataOutput = imageOut.getDataType();

		out.print("\tpublic static void convolve(Kernel2D_"+typeKernel+" kernel, "+typeInput+" input, "+typeOutput+" output , int border ) {\n" +
				"\t\tfinal "+dataOutput+"[] dataDst = output.data;\n" +
				"\t\tfinal "+dataKernel+"[] dataKer = kernel.data;\n" +
				"\n" +
				"\t\tfinal int radius = kernel.getRadius();\n"+
				"\t\tfinal int width = output.getWidth();\n" +
				"\t\tfinal int height = output.getHeight();\n" +
				"\n" +
				"\t\t// convolve along the left and right borders\n" +
				"\t\tfor (int y = 0; y < height; y++) {\n" +
				"\t\t\tint indexDest = output.startIndex + y * output.stride;\n" +
				"\n" +
				"\t\t\tfor ( int x = 0; x < border; x++ ) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tint indexKer = 0;\n" +
				"\t\t\t\tfor( int i = -radius; i <= radius; i++ ) {\n" +
				"\t\t\t\t\tfor (int j = -radius; j <= radius; j++) {\n" +
				"\t\t\t\t\t\ttotal += input.get(x+j,y+i) * dataKer[indexKer++];\n" +
				"\t\t\t\t\t}\n" +
				"\t\t\t\t}\n" +
				"\t\t\t\tdataDst[indexDest++] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\n" +
				"\t\t\tindexDest = output.startIndex + y * output.stride + width-border;\n" +
				"\t\t\tfor ( int x = width-border; x < width; x++ ) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tint indexKer = 0;\n" +
				"\t\t\t\tfor( int i = -radius; i <= radius; i++ ) {\n" +
				"\t\t\t\t\tfor (int j = -radius; j <= radius; j++) {\n" +
				"\t\t\t\t\t\ttotal += input.get(x+j,y+i) * dataKer[indexKer++];\n" +
				"\t\t\t\t\t}\n" +
				"\t\t\t\t}\n" +
				"\t\t\t\tdataDst[indexDest++] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\t\t}\n" +
				"\n" +
				"\t\t// convolve along the top and bottom borders\n" +
				"\t\tfor ( int x = border; x < width-border; x++ ) {\n" +
				"\t\t\tint indexDest = output.startIndex + x;\n" +
				"\n" +
				"\t\t\tfor (int y = 0; y < border; y++, indexDest += output.stride) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tint indexKer = 0;\n" +
				"\t\t\t\tfor( int i = -radius; i <= radius; i++ ) {\n" +
				"\t\t\t\t\tfor (int j = -radius; j <= radius; j++) {\n" +
				"\t\t\t\t\t\ttotal += input.get(x+j,y+i) * dataKer[indexKer++];\n" +
				"\t\t\t\t\t}\n" +
				"\t\t\t\t}\n" +
				"\t\t\t\tdataDst[indexDest] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\n" +
				"\t\t\tindexDest = output.startIndex + (height-border) * output.stride + x;\n" +
				"\t\t\tfor (int y = height-border; y < height; y++, indexDest += output.stride) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tint indexKer = 0;\n" +
				"\t\t\t\tfor( int i = -radius; i <= radius; i++ ) {\n" +
				"\t\t\t\t\tfor (int j = -radius; j <= radius; j++) {\n" +
				"\t\t\t\t\t\ttotal += input.get(x+j,y+i) * dataKer[indexKer++];\n" +
				"\t\t\t\t\t}\n" +
				"\t\t\t\t}\n" +
				"\t\t\t\tdataDst[indexDest] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\t\t}\n" +
				"\t}\n\n");
	}

	public void addConvolutionBound( String typeCast ) {
		String typeOutput = imageOut.getSingleBandName();
		String dataOutput = imageOut.getDataType();

		out.print("\tpublic static void convolve(Kernel2D_"+typeKernel+" kernel, "+typeInput+" input, "+typeOutput+" output , int border ,\n" +
				"\t\t\t\t\t\t\t\t"+sumType+" minValue , "+sumType+" maxValue ) {\n" +
				"\t\tfinal "+dataOutput+"[] dataDst = output.data;\n" +
				"\t\tfinal "+dataKernel+"[] dataKer = kernel.data;\n" +
				"\n" +
				"\t\tfinal int radius = kernel.getRadius();\n"+
				"\t\tfinal int width = output.getWidth();\n" +
				"\t\tfinal int height = output.getHeight();\n" +
				"\n" +
				"\t\t// convolve along the left and right borders\n" +
				"\t\tfor (int y = 0; y < height; y++) {\n" +
				"\t\t\tint indexDest = output.startIndex + y * output.stride;\n" +
				"\n" +
				"\t\t\tfor ( int x = 0; x < border; x++ ) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tint indexKer = 0;\n" +
				"\t\t\t\tfor( int i = -radius; i <= radius; i++ ) {\n" +
				"\t\t\t\t\tfor (int j = -radius; j <= radius; j++) {\n" +
				"\t\t\t\t\t\ttotal += input.get(x+j,y+i) * dataKer[indexKer++];\n" +
				"\t\t\t\t\t}\n" +
				"\t\t\t\t}\n" +
				"\n" +
				"\t\t\t\tif( total < minValue )\n" +
				"\t\t\t\t\ttotal = minValue;\n" +
				"\t\t\t\telse if( total > maxValue )\n" +
				"\t\t\t\t\ttotal = maxValue;\n" +
				"\n"+
				"\t\t\t\tdataDst[indexDest++] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\n" +
				"\t\t\tindexDest = output.startIndex + y * output.stride + width-border;\n" +
				"\t\t\tfor ( int x = width-border; x < width; x++ ) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tint indexKer = 0;\n" +
				"\t\t\t\tfor( int i = -radius; i <= radius; i++ ) {\n" +
				"\t\t\t\t\tfor (int j = -radius; j <= radius; j++) {\n" +
				"\t\t\t\t\t\ttotal += input.get(x+j,y+i) * dataKer[indexKer++];\n" +
				"\t\t\t\t\t}\n" +
				"\t\t\t\t}\n" +
				"\n" +
				"\t\t\t\tif( total < minValue )\n" +
				"\t\t\t\t\ttotal = minValue;\n" +
				"\t\t\t\telse if( total > maxValue )\n" +
				"\t\t\t\t\ttotal = maxValue;\n" +
				"\n"+
				"\t\t\t\tdataDst[indexDest++] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\t\t}\n" +
				"\n" +
				"\t\t// convolve along the top and bottom borders\n" +
				"\t\tfor ( int x = border; x < width-border; x++ ) {\n" +
				"\t\t\tint indexDest = output.startIndex + x;\n" +
				"\n" +
				"\t\t\tfor (int y = 0; y < border; y++, indexDest += output.stride) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tint indexKer = 0;\n" +
				"\t\t\t\tfor( int i = -radius; i <= radius; i++ ) {\n" +
				"\t\t\t\t\tfor (int j = -radius; j <= radius; j++) {\n" +
				"\t\t\t\t\t\ttotal += input.get(x+j,y+i) * dataKer[indexKer++];\n" +
				"\t\t\t\t\t}\n" +
				"\t\t\t\t}\n" +
				"\n" +
				"\t\t\t\tif( total < minValue )\n" +
				"\t\t\t\t\ttotal = minValue;\n" +
				"\t\t\t\telse if( total > maxValue )\n" +
				"\t\t\t\t\ttotal = maxValue;\n" +
				"\n"+
				"\t\t\t\tdataDst[indexDest] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\n" +
				"\t\t\tindexDest = output.startIndex + (height-border) * output.stride + x;\n" +
				"\t\t\tfor (int y = height-border; y < height; y++, indexDest += output.stride) {\n" +
				"\t\t\t\t"+sumType+" total = 0;\n" +
				"\t\t\t\tint indexKer = 0;\n" +
				"\t\t\t\tfor( int i = -radius; i <= radius; i++ ) {\n" +
				"\t\t\t\t\tfor (int j = -radius; j <= radius; j++) {\n" +
				"\t\t\t\t\t\ttotal += input.get(x+j,y+i) * dataKer[indexKer++];\n" +
				"\t\t\t\t\t}\n" +
				"\t\t\t\t}\n" +
				"\n" +
				"\t\t\t\tif( total < minValue )\n" +
				"\t\t\t\t\ttotal = minValue;\n" +
				"\t\t\t\telse if( total > maxValue )\n" +
				"\t\t\t\t\ttotal = maxValue;\n" +
				"\n"+
				"\t\t\t\tdataDst[indexDest] = "+typeCast+"total;\n" +
				"\t\t\t}\n" +
				"\t\t}\n" +
				"\t}\n\n");
	}

	public static void main( String args[] ) throws FileNotFoundException {
		GenerateConvolveJustBorder_General generator = new GenerateConvolveJustBorder_General();
		generator.createAll();
	}
}
